/*
 * bmp-itouch - Multimedia keyboard control plugin for BMP
 * Copyright (C) 2000-2005 Ville Syrjl <syrjala@sci.fi>
 *                         Bryn Davies <curious@ihug.com.au>
 *                         Jonathan A. Davis <davis@jdhouse.org>
 *                         Jeremy Tan <jeremytan@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <X11/Xlib.h>

#include "i18n.h"
#include <gtk/gtk.h>
#include <bmp/plugin.h>
#include <bmp/beepctrl.h>
#include <bmp/configfile.h>
#include <bmp/configdb.h>

static GtkWindow *window;
static GtkRadioButton *rb_use_gnome, *rb_custom;
static GtkVBox *vbox, *gnome_vbox, *custom_vbox;
static GtkTable *custom_config_table, *vol_settings_table;
static GtkHBox *custom_config[7], *vol_inc_hbox, *vol_dec_hbox;

static GtkLabel *play_label, *stop_label, *prev_track_label, 
			 *next_track_label, *vol_up_label, 
			 *vol_down_label, *vol_mute_label,
			 *vol_inc_percent_label, *vol_dec_percent_label;
			 
static GtkSpinButton *play_sb, *stop_sb, *prev_track_sb,
					*next_track_sb, *vol_up_sb,
					*vol_down_sb, *vol_mute_sb, 
					*vol_inc_percent_sb, *vol_dec_percent_sb;
										
static GtkButton *play_b, *stop_b, *prev_track_b,
				*next_track_b, *vol_up_b,
				*vol_down_b, *vol_mute_b;
				
static GtkFrame *keybindings_frame, *volume_adjustment_frame;
				
static GtkHButtonBox  *button_box;
static GtkButton *ok_b, *cancel_b;
	
static GtkAdjustment *spin_button_adjust;
static GtkAlignment *custom_cfg_align; 

typedef struct {
	gboolean use_gnome;
	gint play;
	gint stop;
	gint next_track;
	gint prev_track;
	
	gint vol_up;
	gint vol_down;
	gint mute;
	
	gint vol_inc;
	gint vol_dec;
} MultimediaKeys;

static gboolean grabbed_gnome = FALSE;
static gboolean grabbed_custom = FALSE;
static gboolean plugin_enabled = FALSE;
static gint timeout_tag = 0;
static gint init_done = 0;

static gboolean gui_loaded = FALSE;


MultimediaKeys mkeys;
MultimediaKeys mkeys_gnome;

static Display *display;
static Window root;

static void init(void);
static void about(void);
static void cleanup(void);
static void config(void);
static void grab_keys_gnome(Display * display, Window root);
static void ungrab_keys_gnome(Display * display, Window root);
static void grab_keys_custom(Display * display, Window root);
static void ungrab_keys_custom(Display * display, Window root);
static void get_gnome_keyboard_config();



/********************* Required API Hooks ***************************/

gchar *plugin_filename = "libKeyBoardPlayFunction.so";

static GeneralPlugin gp = {
	NULL,			/* handle */
	NULL,			/* filename */
	-1,			/* xmms_session */
	"Multimedia Keyboard Play Control",	/* Description */
	init,
	about,
	config,
	cleanup
};

GeneralPlugin *get_gplugin_info(void)
{
	gp.description = g_strdup_printf(_("Multimedia Keyboard Control %s"),
					 VERSION);
	return &gp;
}

/**********************************************************************/


/************************ Support Functions ***************************/

static void load_cfg(void)
{
	ConfigDb *cfgdb;
	cfgdb = bmp_cfg_db_open();
					
	/* 4% default */
	mkeys.vol_inc = 4;
	mkeys.vol_dec = 4;

	mkeys.mute = 0;
	mkeys.vol_down = 0;
	mkeys.vol_up = 0;
	mkeys.play = 0;
	mkeys.stop = 0;
	mkeys.prev_track = 0;
	mkeys.next_track = 0;
	
	bmp_cfg_db_get_bool(cfgdb, "Mkeys", "use_gnome",
							&mkeys.use_gnome);
	
	bmp_cfg_db_get_int(cfgdb, "MKeys", "vol_inc",
			  				&mkeys.vol_inc);
	bmp_cfg_db_get_int(cfgdb, "MKeys", "vol_dec",
			  &mkeys.vol_dec);

	bmp_cfg_db_get_int(cfgdb, "MKeys", "mute",
			  &mkeys.mute);
	bmp_cfg_db_get_int(cfgdb, "MKeys", "vol_down",
			  &mkeys.vol_down);
	bmp_cfg_db_get_int(cfgdb, "MKeys", "vol_up",
			  &mkeys.vol_up);
	bmp_cfg_db_get_int(cfgdb, "MKeys", "play",
			  &mkeys.play);
	bmp_cfg_db_get_int(cfgdb, "MKeys", "stop",
			  &mkeys.stop);
	bmp_cfg_db_get_int(cfgdb, "MKeys", "prev_track",
			  &mkeys.prev_track);
	bmp_cfg_db_get_int(cfgdb, "MKeys", "next_track",
			  &mkeys.next_track);
	bmp_cfg_db_close(cfgdb);		
}

static void save_cfg(void)
{
	ConfigDb *cfgdb;

	cfgdb = bmp_cfg_db_open();
	
	bmp_cfg_db_set_bool(cfgdb, "MKeys", "use_gnome",
				mkeys.use_gnome);

	bmp_cfg_db_set_int(cfgdb, "MKeys", "vol_inc",
			   mkeys.vol_inc);
	bmp_cfg_db_set_int(cfgdb, "MKeys", "vol_dec",
			   mkeys.vol_dec);

	bmp_cfg_db_set_int(cfgdb, "MKeys", "mute", 
			   mkeys.mute);
	bmp_cfg_db_set_int(cfgdb, "MKeys", "vol_down",
			   mkeys.vol_down);
	bmp_cfg_db_set_int(cfgdb, "MKeys", "vol_up", 
			   mkeys.vol_up);
	bmp_cfg_db_set_int(cfgdb, "MKeys", "play", 
			   mkeys.play);
	bmp_cfg_db_set_int(cfgdb, "MKeys", "stop", 
	   		   mkeys.stop);
	bmp_cfg_db_set_int(cfgdb, "MKeys", "prev_track",
			   mkeys.prev_track);
	bmp_cfg_db_set_int(cfgdb, "MKeys", "next_track",
			   mkeys.next_track);
			   
	bmp_cfg_db_close(cfgdb);
}


static void x_init(void)
{
	char *display_name;

	display_name = getenv("DISPLAY");
	if (!display_name) {
		perror(_("No display\n"));
		return;
	}
	display = XOpenDisplay(display_name);
	if (!display) {
		perror(_("Can't open display\n"));
		return;
	}
	root = DefaultRootWindow(display);
	if (!root) {
		perror(_("No root window\n"));
		return;
	}
	init_done = 1;
}

static void ungrab_keys_custom(Display * display, Window root)
{
	if (!grabbed_custom)
		return;
	
	if (mkeys.mute)
		ungrab_key(mkeys.mute, display, root);
	if (mkeys.vol_down)
		ungrab_key(mkeys.vol_down, display, root);
	if (mkeys.vol_up)
		ungrab_key(mkeys.vol_up, display, root);
	if (mkeys.play)
		ungrab_key(mkeys.play, display, root);
	if (mkeys.stop)
		ungrab_key(mkeys.stop, display, root);
	if (mkeys.prev_track)
		ungrab_key(mkeys.prev_track, display, root);
	if (mkeys.next_track)
		ungrab_key(mkeys.next_track, display, root);

	grabbed_custom = FALSE;
}

static void grab_keys_custom(Display * display, Window root)
{
	if (grabbed_gnome)
		ungrab_keys_gnome(display, root);
	if (grabbed_custom)
		ungrab_keys_custom(display, root);
	
	if (mkeys.mute)
		grab_key(mkeys.mute, display, root);
	if (mkeys.vol_down)
		grab_key(mkeys.vol_down, display, root);
	if (mkeys.vol_up)
		grab_key(mkeys.vol_up, display, root);
	if (mkeys.play)
		grab_key(mkeys.play, display, root);
	if (mkeys.stop)
		grab_key(mkeys.stop, display, root);
	if (mkeys.prev_track)
		grab_key(mkeys.prev_track, display, root);
	if (mkeys.next_track)
		grab_key(mkeys.next_track, display, root);
		
	/*if (mkey.mute)
		grab_key(mkey.mute, NULL, display, root);
	if (mkey.vol_down)
		grab_key(mkey.vol_down, NULL, display, root);
	if (mkey.vol_up)
		grab_key(mkey.vol_up, NULL, display, root);
	if (mkey.play)
		grab_key(mkey.play, NULL, display, root);
	if (mkey.stop)
		grab_key(mkey.stop, NULL, display, root);
	if (mkey.prev_track)
		grab_key(mkey.prev_track, NULL, display, root);
	if (mkey.next_track)
		grab_key(mkey.next_track, NULL, display, root);*/

	grabbed_custom = TRUE;
}

static void ungrab_keys_gnome(Display * display, Window root)
{
	if (!grabbed_gnome)
		return;
	
	if (mkeys_gnome.mute)
		ungrab_key(mkeys_gnome.mute, display, root);
	if (mkeys_gnome.vol_down)
		ungrab_key(mkeys_gnome.vol_down, display, root);
	if (mkeys_gnome.vol_up)
		ungrab_key(mkeys_gnome.vol_up, display, root);
	if (mkeys_gnome.play)
		ungrab_key(mkeys_gnome.play, display, root);
	if (mkeys_gnome.stop)
		ungrab_key(mkeys_gnome.stop, display, root);
	if (mkeys_gnome.prev_track)
		ungrab_key(mkeys_gnome.prev_track, display, root);
	if (mkeys_gnome.next_track)
		ungrab_key(mkeys_gnome.next_track, display, root);

	grabbed_gnome = FALSE;
}

static void get_gnome_keyboard_config() 
{	/* gnome has its own volume control. code commented out*/
	//mkeys_gnome.mute = gnome_key_config("XF86AudioMute");
	//mkeys_gnome.vol_down = gnome_key_config("XF86AudioLowerVolume");
	//mkeys_gnome.vol_up = gnome_key_config("XF86AudioRaiseVolume");
	
	mkeys_gnome.play = gnome_key_config("XF86AudioPlay");
	mkeys_gnome.stop = gnome_key_config("XF86AudioStop");
	mkeys_gnome.prev_track = gnome_key_config("XF86AudioPrev");
	mkeys_gnome.next_track = gnome_key_config("XF86AudioNext");
	
}

static void grab_keys_gnome(Display * display, Window root)
{
	if(grabbed_custom) 
		ungrab_keys_custom(display, root);
	if (grabbed_gnome)
		ungrab_keys_gnome(display, root);
		
	get_gnome_keyboard_config();
	
	if (mkeys_gnome.mute)
		grab_key(mkeys_gnome.mute, display, root);
	if (mkeys_gnome.vol_down)
		grab_key(mkeys_gnome.vol_down, display, root);
	if (mkeys_gnome.vol_up)
		grab_key(mkeys_gnome.vol_up, display, root);
	if (mkeys_gnome.play)
		grab_key(mkeys_gnome.play, display, root);
	if (mkeys_gnome.stop)
		grab_key(mkeys_gnome.stop, display, root);
	if (mkeys_gnome.prev_track)
		grab_key(mkeys_gnome.prev_track, display, root);
	if (mkeys_gnome.next_track)
		grab_key(mkeys_gnome.next_track, display, root);
	
	/*if (mkey_gnome.mute)
		grab_key(mkey_gnome.mute, NULL, display, root);
	if (mkey_gnome.vol_down)
		grab_key(mkey_gnome.vol_down, NULL, display, root);
	if (mkey_gnome.vol_up)
		grab_key(mkey_gnome.vol_up, NULL, display, root);
	if (mkey_gnome.play)
		grab_key(mkey_gnome.play, NULL, display, root);
	if (mkey_gnome.stop)
		grab_key(mkey_gnome.stop, NULL, display, root);
	if (mkey_gnome.prev_track)
		grab_key(mkey_gnome.prev_track, NULL, display, root);
	if (mkey_gnome.next_track)
		grab_key(mkey_gnome.next_track, NULL, display, root);*/

	grabbed_gnome = TRUE;
}

static void handle_keyevent(int keycode, MultimediaKeys keys)
{
	static gint volume, old_volume;
	gboolean playing, muted;

	playing = xmms_remote_is_playing(gp.xmms_session);
	volume = xmms_remote_get_main_volume(gp.xmms_session);
	if (volume)
		muted = FALSE;
	else
		muted = TRUE;
		
	if (keycode == keys.mute) {
		if (!muted) {
			old_volume = volume;
			volume = 0;
			muted = TRUE;
		} else {
			volume = old_volume;
			muted = FALSE;
		}

		xmms_remote_set_main_volume(gp.xmms_session, volume);
		return;
	}
	if (keycode == keys.vol_down) {
		if (muted) {
			volume = old_volume;
			old_volume = 0;
			muted = FALSE;
		}

		if ((volume -= keys.vol_dec) < 0)
			volume = 0;
		if (volume != old_volume)
			xmms_remote_set_main_volume(gp.xmms_session,
						    volume);
		old_volume = volume;
		return;
	}
	if (keycode == keys.vol_up) {
		if (muted) {
			volume = old_volume;
			old_volume = 0;
			muted = FALSE;
		}

		if ((volume += keys.vol_inc) > 100)
			volume = 100;
		if (volume != old_volume)
			xmms_remote_set_main_volume(gp.xmms_session,
						    volume);
		old_volume = volume;
		return;
	}
	if (keycode == keys.play) {
		if (!playing)
			xmms_remote_play(gp.xmms_session);
		else
			xmms_remote_pause(gp.xmms_session);
		return;
	}
	if (keycode == keys.stop) {
		if (playing)
			xmms_remote_stop(gp.xmms_session);
		else
			xmms_remote_eject(gp.xmms_session);
		return;
	}
	if (keycode == keys.prev_track) {
		xmms_remote_playlist_prev(gp.xmms_session);
		return;
	}
	if (keycode == keys.next_track) {
		xmms_remote_playlist_next(gp.xmms_session);
		return;
	}
}


static gint timeout_func(gpointer data)
{
	long mask = KeyPressMask;
	XEvent event;

	while (XCheckMaskEvent(display, mask, &event))
	if(mkeys.use_gnome)
		handle_keyevent(event.xkey.keycode, mkeys_gnome);
	else
		handle_keyevent(event.xkey.keycode, mkeys);

	return TRUE;
}



/**********************************************************************/

/************************GUI Signal Handlers **************************/

static void ok_button_pressed(GtkWidget *widget, gpointer data) 
{
	
	mkeys.play = gtk_spin_button_get_value_as_int(play_sb);
	mkeys.stop = gtk_spin_button_get_value_as_int(stop_sb);
	mkeys.prev_track = gtk_spin_button_get_value_as_int(prev_track_sb);
	mkeys.next_track = gtk_spin_button_get_value_as_int(next_track_sb);
	mkeys.vol_up = gtk_spin_button_get_value_as_int(vol_up_sb);
	mkeys.vol_down = gtk_spin_button_get_value_as_int(vol_down_sb);
	mkeys.mute = gtk_spin_button_get_value_as_int(vol_mute_sb);
	mkeys.vol_inc = gtk_spin_button_get_value_as_int(vol_inc_percent_sb);
	mkeys.vol_dec = gtk_spin_button_get_value_as_int(vol_dec_percent_sb);
		
	save_cfg();
	if(plugin_enabled)
	{
		if(mkeys.use_gnome)
			grab_keys_gnome(display, root);
		else
			grab_keys_custom(display, root);
	}
	gui_loaded = TRUE;
	gtk_widget_hide_all(GTK_WIDGET(window));
}

static void cancel_button_pressed(GtkWidget *widget, gpointer data) 
{
	if(plugin_enabled)
	{
		if(mkeys.use_gnome)
			grab_keys_gnome(display, root);
		else
			grab_keys_custom(display, root);
	}
	gui_loaded = TRUE;
	gtk_widget_hide_all(GTK_WIDGET(window));
}

static void gnome_rb_selected()
{
	mkeys.use_gnome = TRUE;
	gtk_widget_set_sensitive(GTK_WIDGET(custom_config_table), FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(vol_settings_table), FALSE);
}

static void custom_rb_selected()
{
	mkeys.use_gnome = FALSE;
	gtk_widget_set_sensitive(GTK_WIDGET(custom_config_table), TRUE);
	gtk_widget_set_sensitive(GTK_WIDGET(vol_settings_table), TRUE);
}

static gboolean using_gnome()
{
	ConfigDb *db = bmp_cfg_db_open();
	bmp_cfg_db_get_bool(db, "MKeys", "use_gnome",
			  &mkeys.use_gnome);
	bmp_cfg_db_close(db);
	
	return mkeys.use_gnome;
}

static void get_keycode(GtkWidget * w)
{
	long mask = KeyPressMask;
	XEvent event;
	
	if (!init_done)
		x_init();

	XGrabKeyboard(display, root, False, GrabModeAsync, GrabModeAsync,
		      CurrentTime);

	XMaskEvent(display, mask, &event);

	XUngrabKeyboard(display, CurrentTime);

	gtk_spin_button_set_value(GTK_SPIN_BUTTON(w), event.xkey.keycode);
}

/**********************************************************************/
/************************GUI Support Functions*************************/
static load_settings_into_gui()
{
	load_cfg();
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(play_sb), mkeys.play);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(stop_sb), mkeys.stop);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(prev_track_sb), mkeys.prev_track);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(next_track_sb), mkeys.next_track);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(vol_up_sb), mkeys.vol_up);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(vol_down_sb), mkeys.vol_down);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(vol_mute_sb), mkeys.mute);
	
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(vol_inc_percent_sb), mkeys.vol_inc);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(vol_dec_percent_sb), mkeys.vol_dec);
									
	/* somehow the mkeys.use_gnome setting got lost on the way here so 
		we read it again */
	ConfigDb *db = bmp_cfg_db_open();
	bmp_cfg_db_get_bool(db, "MKeys", "use_gnome",
			  &mkeys.use_gnome);
	bmp_cfg_db_close(db);
	
	if(using_gnome())
	{
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(rb_use_gnome), TRUE);
		gtk_widget_set_sensitive(GTK_WIDGET(custom_config_table), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(vol_settings_table), FALSE);
	}
	else
	{
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(rb_custom), TRUE);
		gtk_widget_set_sensitive(GTK_WIDGET(custom_config_table), TRUE);
		gtk_widget_set_sensitive(GTK_WIDGET(vol_settings_table), TRUE);
	}
}
/**********************************************************************/
/************************API Functions ********************************/

static void init(void)
{
	x_init();
	timeout_tag = gtk_timeout_add(100, timeout_func, NULL);
	
	load_cfg();
	if(using_gnome())
		grab_keys_gnome(display, root);
	else
		grab_keys_custom(display, root);
	plugin_enabled = TRUE;
}

static void about(void)
{
	static GtkWidget *dialog;

	if (dialog != NULL) {
		gdk_window_raise(dialog->window);
		return;
	}

	dialog = xmms_show_message(_("About Multimedia Keyboard Play Control"),
				   _
				   ("Beep Media Player Multimedia Keyboard Control Plugin\n"
				    "Copyright (C) 2000-2004 Ville Syrjl <syrjala@sci.fi>\n"
				    "                        Bryn Davies <curious@ihug.com.au>\n"
				    "                        Jonathan A. Davis <davis@jdhouse.org>\n"
				    "                        Jeremy Tan <jeremytan@users.sourceforge.net>\n\n"
				    "This program is free software; you can redistribute it and/or modify\n"
				    "it under the terms of the GNU General Public License as published by\n"
				    "the Free Software Foundation; either version 2 of the License, or\n"
				    "(at your option) any later version.\n\n"
				    "This program is distributed in the hope that it will be useful,\n"
				    "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
				    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
				    "GNU General Public License for more details.\n\n"
				    "You should have received a copy of the GNU General Public License\n"
				    "along with this program; if not, write to the Free Software\n"
				    "Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n"
				    "NOTE: \n"
				    "This program is now maintained by Jeremy Tan. Please send all\n"
				    "bug reports to Jeremy Tan <jeremytan@users.sourceforge.net>\n\n"
				    "This software is of BETA quality. As such, no support is provided."),
				   _("Ok"), FALSE, NULL, NULL);

	gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed), &dialog);

}

static void config(void)
{	
	if(grabbed_gnome)
		ungrab_keys_gnome(display, root);
	else if (grabbed_custom)
		ungrab_keys_custom(display, root);		
	
	if(gui_loaded)
	{
		load_settings_into_gui();
		gtk_widget_show_all(GTK_WIDGET(window));
		return;
	}
	
	window = g_object_new(GTK_TYPE_WINDOW,
							"default-height", 600,
							"default-width", 600,
							"border-width", 10,
							"title", "Multimedia Keyboard Configuration",
							"window-position", GTK_WIN_POS_CENTER,
							"resizable", FALSE,
							NULL);
							
	/* Closing the current window will have the same effect as pressing the
		cancel button*/	
							
	g_signal_connect(window, "delete-event",
					 	G_CALLBACK(cancel_button_pressed), NULL);
					 	
	vbox = g_object_new(GTK_TYPE_VBOX, "spacing", 6, NULL);
	gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(vbox));
	
	/* Radio Buttons */
	gnome_vbox = g_object_new(GTK_TYPE_VBOX, "spacing", 6, NULL);
	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(gnome_vbox), FALSE, FALSE, 0);
	rb_use_gnome = g_object_new(GTK_TYPE_RADIO_BUTTON,
								"use-underline", TRUE,
								"active", TRUE,
								"label", "Use GNOME keyboard preferences",
								NULL);
	
					
	gtk_box_pack_start(GTK_BOX(gnome_vbox), GTK_WIDGET(rb_use_gnome), FALSE, FALSE, 0);
	
									
	custom_vbox = g_object_new(GTK_TYPE_VBOX, "spacing", 6, NULL);
	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(custom_vbox), FALSE, FALSE, 0);							
	rb_custom = g_object_new(GTK_TYPE_RADIO_BUTTON,
							 "use-underline", TRUE,
							 "label", "Custom keyboard preferences",
							 "active", FALSE, 
							 NULL);
							 
	gtk_signal_connect(GTK_OBJECT(rb_custom), "toggled", 
						G_CALLBACK(custom_rb_selected), 
						NULL);
	gtk_signal_connect(GTK_OBJECT(rb_use_gnome), "toggled", 
					   	G_CALLBACK(gnome_rb_selected), 
					   	NULL);
	g_object_set(rb_custom, "group", rb_use_gnome, NULL);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(rb_use_gnome), TRUE);
	
	gtk_box_pack_start_defaults(GTK_BOX(custom_vbox), GTK_WIDGET(rb_custom));
	
	
	
	/* Custom settings */
	keybindings_frame = g_object_new(GTK_TYPE_FRAME,
										"label", "Key Bindings",
										"label-xalign", 0.02, 
										NULL);
										
	gtk_box_pack_start(GTK_BOX(custom_vbox), 
						GTK_WIDGET(keybindings_frame), 
						FALSE, FALSE, 5);									
	
	custom_config_table = g_object_new(GTK_TYPE_TABLE,
										"row-spacing", 3,
										"column-spacing", 5,
										NULL);
									   
	gtk_container_add(GTK_CONTAINER(keybindings_frame), GTK_WIDGET(custom_config_table));
	gtk_container_border_width(GTK_CONTAINER(custom_config_table), 10);			
	gint i;						
	for(i = 0; i < 7; i++)
	{
		custom_config[i] = g_object_new(GTK_TYPE_HBOX, "spacing", 3, NULL);
	}
	
	spin_button_adjust = GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 255, 1, 5, 0)); 
														   
	custom_cfg_align = g_object_new(GTK_TYPE_ALIGNMENT,
									"xalign", 0.0,
									NULL);
								
	
	/* Custom Play Setting Hbox items*/
	play_label = g_object_new(GTK_TYPE_LABEL, 
								"label", "Play", 
								"justify", GTK_JUSTIFY_RIGHT,
								NULL);
	play_sb = g_object_new(GTK_TYPE_SPIN_BUTTON,
							"adjustment", 
							gtk_adjustment_new(0, 0, 255, 1, 5, 0),
							"digits", 0,
							"value", 0,
							//right-justify", 1,
							NULL);
	play_b = g_object_new(GTK_TYPE_BUTTON,
							"label", "Grab",
							NULL);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(play_sb), TRUE);
	gtk_box_pack_end(GTK_BOX(custom_config[0]), GTK_WIDGET(play_b), FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(custom_config[0]), GTK_WIDGET(play_sb), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(custom_config[0]), GTK_WIDGET(play_label), FALSE, FALSE, 0);
	gtk_table_attach_defaults(custom_config_table, 
					 	GTK_WIDGET(custom_config[0]), 
					 	0, 1, 0, 1);
					 	
	gtk_signal_connect_object(GTK_OBJECT(play_b), "clicked",
				  GTK_SIGNAL_FUNC(get_keycode),
				  GTK_OBJECT(play_sb));
					
	/* Custom Stop Setting Hbox items*/
	stop_label = g_object_new(GTK_TYPE_LABEL, 
								"label", "Stop", 
								"justify", GTK_JUSTIFY_RIGHT,
								NULL);
	stop_sb = g_object_new(GTK_TYPE_SPIN_BUTTON,
							"adjustment", 
							gtk_adjustment_new(0, 0, 255, 1, 5, 0),
							"digits", 0,
							"value", 0,
							NULL);
	stop_b = g_object_new(GTK_TYPE_BUTTON,
							"label", "Grab",
							NULL);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(stop_sb), TRUE);
	gtk_box_pack_end(GTK_BOX(custom_config[1]), GTK_WIDGET(stop_b), FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(custom_config[1]), GTK_WIDGET(stop_sb), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(custom_config[1]), GTK_WIDGET(stop_label), FALSE, FALSE, 0);
	gtk_table_attach_defaults(custom_config_table, 
	                 	GTK_WIDGET(custom_config[1]), 
	                 	1, 2, 0, 1);
	                 	
	gtk_signal_connect_object(GTK_OBJECT(stop_b), "clicked",
				  GTK_SIGNAL_FUNC(get_keycode),
				  GTK_OBJECT(stop_sb));
	
	/* Custom Previous Track Setting Hbox items*/
	prev_track_label = g_object_new(GTK_TYPE_LABEL, 
								"label", "Previous Track", 
								"justify", GTK_JUSTIFY_RIGHT,
								NULL);
	prev_track_sb = g_object_new(GTK_TYPE_SPIN_BUTTON,
							"adjustment", 
							gtk_adjustment_new(0, 0, 255, 1, 5, 0),
							"digits", 0,
							"value", 0,
							NULL);
	prev_track_b = g_object_new(GTK_TYPE_BUTTON,
							"label", "Grab",
							NULL);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(prev_track_sb), TRUE);
	gtk_box_pack_end(GTK_BOX(custom_config[2]), GTK_WIDGET(prev_track_b), FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(custom_config[2]), GTK_WIDGET(prev_track_sb), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(custom_config[2]), GTK_WIDGET(prev_track_label), FALSE, FALSE, 0);
	gtk_table_attach_defaults(custom_config_table, 
					 	GTK_WIDGET(custom_config[2]), 
						0, 1, 1, 2);
	gtk_signal_connect_object(GTK_OBJECT(prev_track_b), "clicked",
				  GTK_SIGNAL_FUNC(get_keycode),
				  GTK_OBJECT(prev_track_sb));
	
	/* Custom Next Track Hbox items*/
	next_track_label = g_object_new(GTK_TYPE_LABEL, 
								"label", "Next Track", 
								"justify", GTK_JUSTIFY_RIGHT,
								NULL);
	next_track_sb = g_object_new(GTK_TYPE_SPIN_BUTTON,
							"adjustment", 
							gtk_adjustment_new(0, 0, 255, 1, 5, 0),
							"digits", 0,
							"value", 0,
							NULL);
	next_track_b = g_object_new(GTK_TYPE_BUTTON,
							"label", "Grab",
							NULL);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(next_track_sb), TRUE);
	gtk_box_pack_end(GTK_BOX(custom_config[3]), GTK_WIDGET(next_track_b), FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(custom_config[3]), GTK_WIDGET(next_track_sb), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(custom_config[3]), GTK_WIDGET(next_track_label), FALSE, FALSE, 0);
	gtk_table_attach_defaults(custom_config_table, 
					GTK_WIDGET(custom_config[3]), 
					1, 2, 1, 2);
	gtk_signal_connect_object(GTK_OBJECT(next_track_b), "clicked",
				  GTK_SIGNAL_FUNC(get_keycode),
				  GTK_OBJECT(next_track_sb));
	
	/* Custom Volume up Setting Hbox items*/
	vol_up_label = g_object_new(GTK_TYPE_LABEL, 
								"label", "Volume Up", 
								//"justify", GTK_JUSTIFY_RIGHT,
								NULL);
	vol_up_sb = g_object_new(GTK_TYPE_SPIN_BUTTON,
							"adjustment", 
							gtk_adjustment_new(0, 0, 255, 1, 5, 0),
							"digits", 0,
							"value", 0,
							NULL);
	vol_up_b = g_object_new(GTK_TYPE_BUTTON,
							"label", "Grab",
							NULL);

	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(vol_up_sb), TRUE);
	gtk_box_pack_end(GTK_BOX(custom_config[4]), GTK_WIDGET(vol_up_b), FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(custom_config[4]), GTK_WIDGET(vol_up_sb), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(custom_config[4]), GTK_WIDGET(vol_up_label), FALSE, FALSE, 0);
	gtk_table_attach_defaults(custom_config_table, 
					GTK_WIDGET(custom_config[4]), 
					0, 1, 2, 3);
	gtk_signal_connect_object(GTK_OBJECT(vol_up_b), "clicked",
				  GTK_SIGNAL_FUNC(get_keycode),
				  GTK_OBJECT(vol_up_sb));
	
	/* Custom Volume Down Setting Hbox items*/
	vol_down_label = g_object_new(GTK_TYPE_LABEL, 
								"label", "Volume Down", 
								"justify", GTK_JUSTIFY_RIGHT,
								NULL);
	vol_down_sb = g_object_new(GTK_TYPE_SPIN_BUTTON,
							"adjustment", 
							gtk_adjustment_new(0, 0, 255, 1, 5, 0),
							"digits", 0,
							"value", 0,
							NULL);
	vol_down_b = g_object_new(GTK_TYPE_BUTTON,
							"label", "Grab",
							NULL);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(vol_down_sb), TRUE);
	gtk_box_pack_end(GTK_BOX(custom_config[5]), GTK_WIDGET(vol_down_b), FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(custom_config[5]), GTK_WIDGET(vol_down_sb), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(custom_config[5]), GTK_WIDGET(vol_down_label), FALSE, FALSE, 0);
	gtk_table_attach_defaults(custom_config_table, 
						GTK_WIDGET(custom_config[5]), 
						1, 2, 2, 3);
						
	gtk_signal_connect_object(GTK_OBJECT(vol_down_b), "clicked",
				  GTK_SIGNAL_FUNC(get_keycode),
				  GTK_OBJECT(vol_down_sb));
	
	/* Custom Volume Mute Setting Hbox items*/
	vol_mute_label = g_object_new(GTK_TYPE_LABEL, 
								"label", "Mute", 
								"justify", GTK_JUSTIFY_RIGHT,
								NULL);
	vol_mute_sb = g_object_new(GTK_TYPE_SPIN_BUTTON,
							"adjustment", 
							gtk_adjustment_new(0, 0, 255, 1, 5, 0),
							"digits", 0,
							"value", 0,
							NULL);
	vol_mute_b = g_object_new(GTK_TYPE_BUTTON,
							"label", "Grab",
							NULL);
	gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(vol_mute_sb), TRUE);
	gtk_box_pack_end(GTK_BOX(custom_config[6]), GTK_WIDGET(vol_mute_b), FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(custom_config[6]), GTK_WIDGET(vol_mute_sb), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(custom_config[6]), GTK_WIDGET(vol_mute_label), FALSE, FALSE, 0);
	gtk_table_attach_defaults(custom_config_table, 
						GTK_WIDGET(custom_config[6]), 
						0, 1, 3, 4);
		gtk_signal_connect_object(GTK_OBJECT(vol_mute_b), "clicked",
				  GTK_SIGNAL_FUNC(get_keycode),
				  GTK_OBJECT(vol_mute_sb));
	
					 	
	/* Volume increment/decrement settings */
	volume_adjustment_frame  = g_object_new(GTK_TYPE_FRAME, 
										"label", "Volume Adjustment",
										"label-xalign", 0.02, 
										NULL);
	gtk_box_pack_start(GTK_BOX(vbox), 
						GTK_WIDGET(volume_adjustment_frame), 
						FALSE, FALSE, 5);
						
	vol_settings_table = g_object_new(GTK_TYPE_TABLE,
										"row-spacing", 3,
										NULL);
	gtk_container_border_width(GTK_CONTAINER(vol_settings_table), 10);									
	gtk_container_add(GTK_CONTAINER(volume_adjustment_frame), GTK_WIDGET(vol_settings_table));
	
	vol_inc_hbox = g_object_new(GTK_TYPE_HBOX, 
									"spacing", 0,
									//"padding", 10, 
									"homogeneous", TRUE,
									NULL);
	gtk_table_attach_defaults(vol_settings_table,
						GTK_WIDGET(vol_inc_hbox),
						1, 2, 0, 1);
	vol_inc_percent_label = g_object_new(GTK_TYPE_LABEL, 
										 	"label", "Volume Increment",
										 	NULL);
	vol_inc_percent_sb = g_object_new(GTK_TYPE_SPIN_BUTTON,
										"adjustment",
										gtk_adjustment_new(0, 0, 100, 1, 5, 0),
										"digits", 0,
										"value", mkeys.vol_inc,
										NULL);
		
	gtk_box_pack_start(GTK_BOX(vol_inc_hbox),
								GTK_WIDGET(vol_inc_percent_label),
								FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vol_inc_hbox),
								GTK_WIDGET(gtk_label_new("(%)")),
								FALSE, FALSE, 10);
	gtk_box_pack_start(GTK_BOX(vol_inc_hbox),
								GTK_WIDGET(vol_inc_percent_sb),
								TRUE, FALSE, 0);
								
	vol_dec_hbox = g_object_new(GTK_TYPE_HBOX, 
									"spacing", 0, 
									"homogeneous", TRUE,
									NULL);
	gtk_table_attach_defaults(vol_settings_table,
						GTK_WIDGET(vol_dec_hbox),
						1, 2, 1, 2);
	vol_dec_percent_label = g_object_new(GTK_TYPE_LABEL, 
										 	"label", "Volume Decrement",
										 	NULL);
	vol_dec_percent_sb = g_object_new(GTK_TYPE_SPIN_BUTTON,
										"adjustment",
										gtk_adjustment_new(0, 0, 100, 1, 5, 0),
										"digits", 0,
										"value", mkeys.vol_dec,
										NULL);
	
	gtk_box_pack_start(GTK_BOX(vol_dec_hbox),
								GTK_WIDGET(vol_dec_percent_label),
								FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vol_dec_hbox),
								GTK_WIDGET(gtk_label_new("(%)")),
								FALSE, FALSE, 10);							
	gtk_box_pack_start(GTK_BOX(vol_dec_hbox),
								GTK_WIDGET(vol_dec_percent_sb),
								TRUE, FALSE, 5);
		
	/* Button bar at bottom */
	button_box = g_object_new(GTK_TYPE_HBUTTON_BOX, 
								"layout-style", GTK_BUTTONBOX_END, 
								NULL);
	gtk_box_pack_end(GTK_BOX(vbox), GTK_WIDGET(button_box), FALSE, FALSE, 0);
	
	ok_b = g_object_new(GTK_TYPE_BUTTON,
						"use-stock", TRUE,
						"label", GTK_STOCK_OK,
						NULL);
	g_signal_connect(ok_b, "clicked", G_CALLBACK(ok_button_pressed), NULL);
	
	cancel_b = g_object_new(GTK_TYPE_BUTTON,
						"use-stock", TRUE,
						"label", GTK_STOCK_CANCEL,
						NULL);
	g_signal_connect(cancel_b, "clicked", G_CALLBACK(cancel_button_pressed), NULL);
	
	gtk_box_pack_end(GTK_BOX(button_box), GTK_WIDGET(cancel_b), FALSE, FALSE, 0);
	gtk_box_pack_end(GTK_BOX(button_box), GTK_WIDGET(ok_b), FALSE, FALSE, 0);
	

	load_settings_into_gui();
	gtk_widget_show_all(GTK_WIDGET(window));
}

static void cleanup(void)
{
	if(mkeys.use_gnome)
		ungrab_keys_gnome(display, root);
	else
		ungrab_keys_custom(display, root);
		
	if (timeout_tag)
		gtk_timeout_remove(timeout_tag);
	timeout_tag = 0;
	XCloseDisplay(display);
	plugin_enabled = FALSE;
}

